/**
 * Accumulate and render point cloud to an image plane
 */
//#define USE_MT_NODE_HANDLE
#include <boost/bind.hpp>
#include <boost/format.hpp>
#include <cv_bridge/cv_bridge.h>
#include <image_transport/image_transport.h>
#include <limits>
#include <nodelet/nodelet.h>
#include <opencv2/opencv.hpp>
#include <pcl/io/io.h>
#include <pluginlib/class_list_macros.h>
#include <ros/ros.h>
#include <sensor_msgs/CameraInfo.h>
#include <sensor_msgs/image_encodings.h>
#include <sensor_msgs/PointCloud2.h>
#include <tf/transform_datatypes.h>
#include <tf/transform_listener.h>

#include <deque>          // std::deque
#include <list>           // std::list

#include <boost/shared_ptr.hpp>

#define NO_EXTTYPE
#include <stdio.h>
#include "dynamic/num_array.h"

#include <QtGui/QApplication>
#include <QtGui/QLabel>
#include <QtCore/QBasicTimer>
#include <QtGui/QPainter>
#include <QtGui/QDialog>
#include <QtGui/QMainWindow>
//#include <QtGui/QMessageBox>
#include <QtGui/QKeyEvent>
#include <QtGui/QMouseEvent>
//
//#include <GL/glut.h>

//#include "GL/gl.h"
#include "GL/glew.h"
#include "GL/freeglut.h"
#include <GL/glut.h>

#include "geom/nmmatrix.h"
#include <math.h>
#include "../../thermo_cam/src/IronPalette.h"
#include "debug/performance.h"

using namespace exttype;

int getFieldIndex(const sensor_msgs::PointCloud2 &cloud, const std::string fieldName) {
    for(size_t i = 0; i < cloud.fields.size(); i++) {
        if(cloud.fields[i].name == fieldName) {
            return i;
        }
    }
    return -1;
}

cv::Mat rotationFromTransform(const tf::Transform &t) {
    return (cv::Mat_<double>(3, 3) << t.getBasis()[0][0], t.getBasis()[0][1], t.getBasis()[0][2],
            t.getBasis()[1][0], t.getBasis()[1][1], t.getBasis()[1][2],
            t.getBasis()[2][0], t.getBasis()[2][1], t.getBasis()[2][2]);
}

cv::Mat translationFromTransform(const tf::Transform &t) {
    return (cv::Mat_<double>(3, 1) << t.getOrigin()[0], t.getOrigin()[1], t.getOrigin()[2]);
}

cv::Mat matFromCloud(sensor_msgs::PointCloud2ConstPtr cloud, const std::string fieldName, const int numElements) {
    const int numPoints = cloud->width * cloud->height;
    int fieldIndex = getFieldIndex(*cloud, fieldName);
    void *data = const_cast<void *>(static_cast<const void *>(&cloud->data[cloud->fields[fieldIndex].offset]));
    int matType;
    switch(cloud->fields[fieldIndex].datatype) {
    case sensor_msgs::PointField::FLOAT32:
        matType = CV_32FC1;
        break;
    case sensor_msgs::PointField::FLOAT64:
        matType = CV_64FC1;
        break;
    case sensor_msgs::PointField::UINT8:
        matType = CV_8UC1;
        break;
    default:
        assert(0);
        break;
    }
    return cv::Mat(numPoints, numElements, matType, data, cloud->point_step);
};

void glcheck() {
    GLenum errCode;
    const GLubyte *errString;
    if((errCode = glGetError()) != GL_NO_ERROR) {
        errString = gluErrorString(errCode);
        fprintf(stderr, "OpenGL Error: %s, code %x\n", errString,errCode);
        fflush(stdout);
        exit(1);
    };
};

//_______________________________________________________________________________________

class PointCloudImage : public nodelet::Nodelet {
public:
    bool display;
    std::string fixedFrame;
    std::string camera_topic;
    int pointCloudQueueSize;
    double waitForTransform;
    int count;
public:
    PointCloudImage();
    virtual ~PointCloudImage();
    void clear();
    void onInit();
private:
    QLabel * label;
    QImage img;
private:
    int Width;
    int Height;
    bool laser_empty;
    ros::NodeHandle nh;
    ros::NodeHandle pnh;
    image_transport::ImageTransport * it;//(nh);
private:
    GLuint fbo, d_render_buf;
    static const float zFar;  //100 m
    static const float zNear; //10 cm
private:
    tf::TransformListener transformListener;
    ros::Subscriber cameraSubscriber;
    image_transport::Publisher imagePublisher;
    ros::Subscriber pointCloudSub;
    dynamic::num_array<float,2> laser_image;
    dynamic::num_array<short unsigned int,2> ilaser_image;
    ros::Time last_scan_t;
    //
    struct p_cloud {
        //cv::Mat X;
        //cv::Mat W;
    };
    std::deque<cv::Mat> w_clouds; //! poit clouds mapped into fixed_frame
    typedef std::deque<cv::Mat>::iterator w_clouds_it;
    //
private:
    //double maxCloudAge;
    void init_image(int width, int height);
    void pointCloudCallback(const sensor_msgs::PointCloud2ConstPtr &cloudMsg);
    void cameraInfoCallback(const sensor_msgs::CameraInfoConstPtr &cameraInfoMsg);
    void laser_render(const sensor_msgs::CameraInfoConstPtr &cameraInfoMsg, const tf::StampedTransform & tf1);
    void laser_display();
    void laser_publish(const sensor_msgs::CameraInfoConstPtr &cameraInfoMsg);
};

const float PointCloudImage::zFar = 100;  //100 m
const float PointCloudImage::zNear = 0.1; //10 cm

PointCloudImage::PointCloudImage() :
    nh(),
    pnh("~"),
    transformListener(ros::Duration(200.0),true),
    fixedFrame("/odom"),
    pointCloudQueueSize(300),
    waitForTransform(0.3),
    label(0),
    Width(0),Height(0),count(0),it(0) {
}

PointCloudImage::~PointCloudImage() {
    clear();
};

void PointCloudImage::clear() {
    if(label) {
        delete label;
    };
    glDeleteFramebuffers(1,&fbo);
    glDeleteRenderbuffers(1,&d_render_buf);
};

void PointCloudImage::onInit() {
    NODELET_INFO("PointCloudImage::onInit: Initializing...");
    // Get and process parameters.
    pnh.param("fixed_frame", fixedFrame, fixedFrame);
    pnh.param("point_cloud_queue_size", pointCloudQueueSize, pointCloudQueueSize);
    pointCloudQueueSize = pointCloudQueueSize >= 1 ? pointCloudQueueSize : 1;
    pnh.param("wait_for_transform", waitForTransform, waitForTransform);
    waitForTransform = waitForTransform >= 0.0 ? waitForTransform : 0.0;
    pnh.param("camera_topic", camera_topic, std::string("openni_camera"));
    pnh.param("display", display, true);
    std::string pclInTopic;
    pnh.param("scan_point_cloud", pclInTopic, std::string("scan_point_cloud"));
    //
    cameraSubscriber = nh.subscribe(camera_topic+"/camera_info", 2, &PointCloudImage::cameraInfoCallback,this);
    std::cout<<"Subscribed "<<camera_topic+"/camera_info"<<"\n";
    fflush(stdout);
    pointCloudSub = nh.subscribe<sensor_msgs::PointCloud2>(pclInTopic, 50, &PointCloudImage::pointCloudCallback, this);
    std::cout<<"Subscribed "<<pclInTopic<<"\n";
    fflush(stdout);
    it = new image_transport::ImageTransport(nh);
    imagePublisher = it->advertise(camera_topic+"/laser_image", 5);
    std::cout<<"Advertised "<<camera_topic+"/laser_image"<<"\n";
    fflush(stdout);
}

void PointCloudImage::init_image(int width, int height) {
    Width = width;
    Height = height;
    if(Width>512) {
        Width = width/2;
        Height = height/2;
    };
    laser_image.resize(exttype::mint2(Width,Height));
    img = QImage(Width, Height, QImage::Format_RGB32);
    label = new QLabel();
    if(display) {
        label->show();
    } else {
        label->setVisible(false);
    }
    label->setGeometry(100,100,10,10);
    label->setScaledContents(true);
    if(Width<400) {
        label->resize(2*Width,2*Height);
    } else {
        label->resize(Width,Height);
    };
    //
    glutInitWindowSize(Width,Height);
    //glutInitWindowSize(Width/2,Height/2);
    int argc1=0;
    char *argv1[0];
    glutInit(&argc1, argv1);
    glutCreateWindow("GLEW window of the RIGHT size");
    glutHideWindow();
    glewExperimental=GL_TRUE;
    GLenum err = glewInit();
    if(GLEW_OK != err) {
        /* Problem: glewInit failed, something is seriously wrong. */
        fprintf(stderr, "Error: %s\n", glewGetErrorString(err));
        exit(1);
    };
    printf("Vendor: %s\n", glGetString(GL_VENDOR));
    printf("Renderer: %s\n", glGetString(GL_RENDERER));
    printf("Version: %s\n", glGetString(GL_VERSION));
    fflush(stdout);
    //
    glGenFramebuffers(1,&fbo);
    glGenRenderbuffers(1,&d_render_buf);
    glBindFramebuffer(GL_FRAMEBUFFER,fbo);
    //
    glBindRenderbuffer(GL_RENDERBUFFER,d_render_buf);
    glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, Width, Height);
    glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, d_render_buf);
    //
    GLenum e = glCheckFramebufferStatus(GL_FRAMEBUFFER);
    if(e!=GL_FRAMEBUFFER_COMPLETE) {
        //perror("Error: GL FrameBuffer, ");
        GLenum errCode;
        const GLubyte *errString;

        if((errCode = glGetError()) != GL_NO_ERROR) {
            errString = gluErrorString(errCode);
            fprintf(stderr, "OpenGL Error: %s, %x, %x\n", errString,e,errCode);

            fflush(stdout);
        }
//       std::cout<<gluErrorString(e)<<"\n";
        exit(1);
    };
//
    glEnable(GL_DEPTH_TEST);
    //glcheck();
    //glClear(GL_DEPTH_BUFFER_BIT);
    //glcheck();
    glDisable(GL_STENCIL_TEST);
    glDisable(GL_CULL_FACE);
    glcheck();
};


void PointCloudImage::laser_render(const sensor_msgs::CameraInfoConstPtr &cameraInfoMsg, const tf::StampedTransform & tf1) {
    glBindFramebuffer(GL_FRAMEBUFFER,fbo);
    glClear(GL_DEPTH_BUFFER_BIT);
    {
        // set projection matrix
        geom::matrixn<4> m;
        m << 0;
        m(0,0) = cameraInfoMsg->K[0]/Width*2;
        m(0,1) = cameraInfoMsg->K[1]/Width*2;
        m(1,0) = cameraInfoMsg->K[3]/Height*2;
        m(1,1) = cameraInfoMsg->K[4]/Height*2;
        m(3,3) = 1;
        m(2,2) = 1;
        glMatrixMode(GL_PROJECTION);
        glLoadMatrixd(&m(0,0));
        glFrustum(-zNear,zNear,-zNear,zNear,zNear,zFar);
    }
    {
        //set transformation matrix to current camera
        double m[16];
        tf1.getOpenGLMatrix(m);
        glMatrixMode(GL_MODELVIEW);
        for(int j=0; j<4; ++j)m[2+j*4] = -m[2+j*4]; // invert z axis
        glLoadMatrixd(m);
    }
    //glPointSize(5.0);
    glBegin(GL_QUADS);
    //
    for(w_clouds_it pcloud=w_clouds.begin(); pcloud!=w_clouds.end(); ++pcloud) {
        cv::Mat & X = *pcloud;
        float * pm = &X.at<float>(0,0);
        for(int i=0; i<X.rows; ++i) {
            for(int j=0; j<4; ++j) {
                glVertex3fv(pm);
                pm+=3;
            };
        };
    };
    glEnd();
    glFinish();
    //glcheck();
    glReadPixels(0,0,Width,Height,GL_DEPTH_COMPONENT,GL_FLOAT,&laser_image[0]);
    //glcheck();
    // convert normalized depth coordinates to the true depth
    float Q  = zFar / ( zFar - zNear );
    float inv_Zfar = 1.0 / zFar;
    //laser_image*=zFar;

//    min_depth = zFar;
//    max_depth = zNear;

    ilaser_image.resize(laser_image.size());

    for(int i=0;i<laser_image.length();++i){
        float z_b = laser_image[i];
        if(z_b==1){ // this pixel is empty
            laser_image[i] = 0;
        }else{
            float z_n = 2.0 * z_b - 1.0;
            float z_e = 2.0 * zNear * zFar / (zFar + zNear - z_n * (zFar - zNear));
            laser_image[i] = z_e;
        };
        ilaser_image[i] = (unsigned short int)( std::min(1000*laser_image[i],(float)std::numeric_limits<unsigned short int>::max()));
    };
};

void PointCloudImage::laser_display() {
    if(display && !laser_empty) {
        //float min_depth = std::max(float(0.1),laser_image.min().first);
        float max_depth = zNear;
        float min_depth = zFar;
        for(int i = 0; i < (Width) *(Height); ++i) {
            float d = laser_image[i];
            if(d <zFar && max_depth < d) {
                max_depth = d;
            };
            if(d > 0 && min_depth > d) {
                min_depth = d;
            };
        };
        //max_depth = max_depth+0.1;
        //fflush(stdout);
        //char s [1024];
        //sprintf(s,"depth [%5.3f,%5.3f]\n",min_depth,max_depth);
        //std::cout<<s;
        //std::cout.flush();
        //
        QRgb * pimg = (QRgb*)img.bits();
        for(int i = 0; i < (Width) *(Height); ++i) {
            //float * pz = &laser_image[i];
            //double V = 1.0;
            //pimg[i] = qRgb(pixel[0],pixel[1],pixel[2]);
            pimg[i] = qRgb(0,0,0);
            float d = laser_image[i];
            if(d<zFar && d>0) {
                double v = 1/(double(d));
                double V = (v-1/max_depth)/(1/min_depth-1/max_depth);
                //pimg[i] = qRgb(255-floor((V)*255),floor((V)*255),0);
                //int a = floor((V)*239);
                //pimg[i] = qRgb(Iron[a][0],Iron[a][1],Iron[a][2]);
                int a = floor((V*0.8+0.2)*255);
                int b = floor((0.5-std::min(V,0.5))*255);
                pimg[i] = qRgb(a,a,(255-a)/2);
            };
        };
        label->setPixmap(QPixmap::fromImage(img));
    };
};

void PointCloudImage::laser_publish(const sensor_msgs::CameraInfoConstPtr &cameraInfoMsg) {
    sensor_msgs::ImagePtr msg(new sensor_msgs::Image());
    msg->header = cameraInfoMsg->header;
    //msg->encoding = sensor_msgs::image_encodings::TYPE_32FC1;
    msg->encoding = sensor_msgs::image_encodings::MONO16;
    if(laser_empty) {
        msg->data.resize(0);
        msg->step = 0;
        msg->height = 0;
        msg->width = 0;
    } else {
        int num_bytes = ilaser_image.byte_size();
        msg->data.resize(num_bytes);
        memcpy(&msg->data[0],ilaser_image.begin(),num_bytes);
        msg->step = ilaser_image.byte_stride(0,1);
        msg->height = ilaser_image.size()[1];
        msg->width = ilaser_image.size()[0];
    };
    //msg->header.seq = count;
    imagePublisher.publish(msg);
};


void PointCloudImage::cameraInfoCallback(const sensor_msgs::CameraInfoConstPtr &cameraInfoMsg) {
    debug::PerformanceCounter c1;
    //
    laser_empty = false;
    ++count;
    if(Width==0) {
        init_image(cameraInfoMsg->width,cameraInfoMsg->height);
    };

    std::string err;
    //! wait for the tf cloud_frame -> fixed_frame
    transformListener.waitForTransform(cameraInfoMsg->header.frame_id, cameraInfoMsg->header.stamp, // target frame and time
                                       fixedFrame, cameraInfoMsg->header.stamp, // source frame and time
                                       fixedFrame, ros::Duration(0.01)); //    this is time critical, dont wait

    tf::StampedTransform tf1;
    try {
        transformListener.lookupTransform(cameraInfoMsg->header.frame_id, cameraInfoMsg->header.stamp, // target frame and time
                                          fixedFrame, cameraInfoMsg->header.stamp, // source frame and time
                                          fixedFrame, tf1);
    } catch(tf::TransformException ex) {
        try {
            transformListener.lookupTransform(cameraInfoMsg->header.frame_id, cameraInfoMsg->header.stamp-ros::Duration(0.1), // target frame and time
                                              fixedFrame, cameraInfoMsg->header.stamp-ros::Duration(0.1), // source frame and time
                                              fixedFrame, tf1);
        } catch(tf::TransformException ex) {

            std::cout<<"skipping image frame -- no tf "<<fixedFrame<<"->"<<cameraInfoMsg->header.frame_id<<"\n";
            laser_empty = true;
        };
    };

    debug::PerformanceCounter c2;
    c2.pause();
    if(!laser_empty && !w_clouds.empty()) {
        c2.start();
        laser_render(cameraInfoMsg,tf1);
        c2.stop();
        laser_display();
    };
    c1.stop();
    laser_publish(cameraInfoMsg);

    //float max_depth = laser_image.max().first;
    //std::cout<<"["<<min_depth<<","<<max_depth<<"]\n";
    //display image
    char s[1024];
    sprintf(s,"frame time %4.2f (render %4.2f) ",c1.time(),c2.time());
    std::cout<<s;
    fflush(stdout);
    for(int i=0; i<strlen(s); ++i)std::cout<<"\b";
    //std::cout<<c2.time()<<" | ";
    //std::cout<<c1.time()<<"\n";
    //fflush(stdout);
}

void PointCloudImage::pointCloudCallback(const sensor_msgs::PointCloud2ConstPtr &cloudMsg) {


    //std::cout<<cloudMsg->width<<"x"<<cloudMsg->height<<"\n";fflush(stdout);
    double dt;
    if(w_clouds.size()==0) { //first scan
        last_scan_t = cloudMsg->header.stamp;
        dt = 0.02;
    } else {
        dt = (cloudMsg->header.stamp - last_scan_t).toSec();
        if(dt<0 || dt>0.5) {
            dt = 0.02;
        };
        last_scan_t = cloudMsg->header.stamp;
    };
    //std::cout<<dt<<"\n";

    const int numPoints = cloudMsg->width * cloudMsg->height;
    const std::string cloudFrame = cloudMsg->header.frame_id;
    if(numPoints == 0 || cloudMsg->data.size() == 0) {
        return;
    };

    //! wait for the tf cloud_frame -> fixed_frame
    transformListener.waitForTransform(fixedFrame, cloudMsg->header.stamp, // target frame and time
                                       cloudMsg->header.frame_id, cloudMsg->header.stamp, // source frame and time
                                       fixedFrame, ros::Duration(waitForTransform));

    tf::StampedTransform tf1;
    try {
        transformListener.lookupTransform(fixedFrame, cloudMsg->header.stamp, // target frame and time
                                          cloudMsg->header.frame_id, cloudMsg->header.stamp, // source frame and time
                                          fixedFrame, tf1);
    } catch(tf::TransformException ex) {
        std::cout<<"skipping scan -- no tf "<<cloudMsg->header.frame_id<<"->"<<fixedFrame<<"\n";
        return;
    };

    //debug::PerformanceCounter c1;

    if(w_clouds.size() >=  pointCloudQueueSize) {
        w_clouds.pop_back();
    };

    cv::Mat camRotation;
    rotationFromTransform(tf1).convertTo(camRotation,CV_32FC1);
    cv::Mat camTranslation;
    translationFromTransform(tf1).convertTo(camTranslation,CV_32FC1);
    cv::Mat objectPoints;
    matFromCloud(cloudMsg, "x", 3).convertTo(objectPoints, CV_32FC1);

    std::vector<float> d(numPoints);
    std::vector<float> t1(numPoints);
    std::vector<float> t2(numPoints);
    std::vector<float> t3(numPoints);
    for(int i=0; i<numPoints; ++i) {
        float x = objectPoints.at<float>(i,0);
        float y = objectPoints.at<float>(i,1);
        float r = sqrt(x*x+y*y);
        t1[i] = y*0.01;
        t2[i] = -x*0.01;
        float sina = y/r;
        //t3[i] = 0.04*sina;
        //d[i] = r;
        //t3[i] = 0.02*r;
        t3[i] = (0.01*r*math::sgn(y)+y*dt);
    };
//
//    std::cout<<objectPoints.rows<<"x"<<objectPoints.cols<<"\n";fflush(stdout);
//    cv::Mat P  = objectPoints(cv::Range(0,5),cv::Range(0,3));
//    std::cout<<P<<"\n";fflush(stdout);
//    exit(1);
    objectPoints = objectPoints * camRotation.t() + cv::repeat(camTranslation.t(), objectPoints.rows, 1);
    objectPoints = objectPoints.reshape(3);

    //cv::Point3D32f dx(1,0,0);


    cv::Mat X(objectPoints.rows,3*4,CV_32FC1);
    for(int i=0; i<objectPoints.rows; ++i) {
        for(int j=0; j<4; ++j) {
            for(int k=0; k<3; ++k) {
                float a = 0;
                //float ts = 0.02;
                //float zs = 0.01;
                float s = -1;
                if(j==1 || j==2) {
                    s = 1;
                };
                a += s*t1[i]*camRotation.at<float>(k,0);
                a += s*t2[i]*camRotation.at<float>(k,1);

                if(j>1) {
                    a += t3[i]*camRotation.at<float>(k,2);
                }
                //a = a + ((j==1 || j==2) & k==0);
                //a = a + ((j>1) & k==1);
                //a = a*0.02*d[i];
                a += objectPoints.at<float>(i,k);
                /*
                 if(j>1){
                    a += camRotation.at<float>(2,k)*0.02*d[i];
                }
                if(j==1 || j==2){
                    a += objectPoints.at<float>(i+1,k);
                }else{
                    a += objectPoints.at<float>(i,k);
                };
                */
                X.at<float>(i,j*3+k) = a;
            };
            //X.at<geom::mfloat3>(i,j*3) = objectPoints.at<geom::mfloat3>(i,0)+geom::mfloat3(j==1 || j==2, j>1,0)*0.01;
        };
    };
    w_clouds.push_front(X);
    //std::cout<<"cloud_t:" <<c1.time()<<"\n";
}

//___________________________________________________________

class myApp : public QApplication {
public:
    QLabel label;
    PointCloudImage node;
    int timer_count;
public:
    myApp(int & argc, char ** argv):QApplication(argc,argv) {
        timer_count = 0;
        node.onInit();
        startTimer(1000/20);
    };
    //
    ~myApp() {
    };

    void timerEvent(QTimerEvent * event) {
        ros::spinOnce();
        ++timer_count;
        if(node.count<1) {
            if(timer_count % 100 ==0) {
                std::cout<<"waiting for inputs\n";
                std::cout.flush();
            };
        };
        //
        if(!ros::ok()) {
            //ros::shutdown();
            printf("quitting");
            fflush(stdout);
            exit(0);
        };
    };
    bool notify(QObject * receiver, QEvent * ev) {
        if(ev->type() == QEvent::MouseButtonPress) {
            QMouseEvent * e = static_cast<QMouseEvent *>(ev);
            if(e->button()==Qt::LeftButton) {
                return true;
            };
        };
        if(ev->type() == QEvent::KeyPress) {
            QKeyEvent * e = static_cast<QKeyEvent *>(ev);
            if(e->key() == Qt::Key_Backspace) {
                return true;
            } else if(e->key() == Qt::Key_Return) {
                return true;
            } else if(e->key() == Qt::Key_Tab) {
                return true;
            } else if(e->key() == Qt::Key_H) {
                node.display = false;
                return true;
            } else if(e->key() == Qt::Key_Escape) {
                exit(0);
                return true;
            };
        };
        return QApplication::notify(receiver,ev);
    };
};


int main(int argc, char **argv) {
    ros::init(argc, argv, "point_cloud_to_image");
    int argc1=0;
    char *argv1[0];
    myApp * a = new myApp(argc1, argv1);
    a->exec();
    delete a;
}

